// -*- c++ -*-
// $Id$
/*
 * Copyright 2005 Spencer Olson
 *
 * $Log$
 *
 */


#include <octave/oct.h>
#include <octave/pager.h>
#include <olson-tools/physical.h>
#include <olson-tools/bfield.h>
#include <olson-tools/Vector.h>

using namespace physical::constants;
using namespace physical::units;

typedef BField::BCalcs<
    BField::AddSrc<
        BField::ThinWireSrc,
        BField::BgSrc
    >
> BFieldSrc;


DEFUN_DLD (bfield, args, nargout,
"  [Bx By Bz (d|B|/dx) (d|B|/dy) (d|B|/dz)] = bfield(wires, x, y, z [, bg, [, delta] ])\n"
"  Evaluate BField magnitude.\n"
"\n"
"Copyright 2005 Spencer Olson:\n"
"\n"
"  This calculates the magnetic field from specified wires at specified locations. :-)\n"
"\n"
"  wires : Array of info for finite length infinitesimal wire segments.\n"
"          Each row consists of the complete information for a wire segment:\n"
"          [ Pa, Pb, I]\n"
"          where Pa and Pb are both 3-(row)vectors indicating the start and stop\n"
"          position of the wire segment.  The last column, I, indicates the\n"
"          magnitude of the current flowing from Pa to Pb.\n"
"  x   : x position to evaluate to\n"
"  y   : y position to evaluate to\n"
"  z   : z position to evaluate to\n"
"  x, y, and z must have the same size\n"
"  \n"
"  bg  :   background field.  This must be a three vector:  [Bgx Bgy Bgz]\n"
"           [Default [0,0,0]]\n"
"  \n"
"  delta :  delta == dx == dy == dz for calculating d|B|/d{x,y,z}\n"
"           [Default 10e-6]\n"
"  \n"
"  Return:  \n"
"       Bx, By, Bz are x, y, and z components of B-Field with sizes == size(x)\n"
) {
    if (args.length() < 4 || args.length() > 6) {
        print_usage("bfield");
        return octave_value();
    }

    if (nargout < 1 || nargout > 6) {
        print_usage("bfield");
        return octave_value();
    }

    Matrix wires (args(0).matrix_value());
    if ( wires.columns() != 7 ) {
        octave_stdout << "wires matrix size '"
                      << wires.rows() << 'X' << wires.columns()
                      << "'\nmust have 7 columns" << std::endl;
        print_usage("bfield");
        return octave_value();
    }

    BFieldSrc bsrc;
    { /* init bsrc. */
        for (int i = 0; i < wires.rows(); i++) {
            bsrc.currents.push_back(
                BField::ThinCurrentElement(
                            wires(i,0),
                            wires(i,1),
                            wires(i,2),
                            wires(i,3),
                            wires(i,4),
                            wires(i,5),
                            wires(i,6))
            );
        }/* for */

        if (args.length() >= 5) {
            ColumnVector bg(args(4).vector_value());
            if (bg.length() != 3) {
                print_usage("bfield");
                return octave_value();
            }

            bsrc.bg[X] = bg(0);
            bsrc.bg[Y] = bg(1);
            bsrc.bg[Z] = bg(2);
        }

        if (args.length() == 6) {
            bsrc.delta = args(5).double_value();
        }
    }

    Matrix x (args(1).matrix_value()),
           y (args(2).matrix_value()),
           z (args(3).matrix_value());

    if ( x.columns() != y.columns() ||
         x.columns() != z.columns() ||
         x.rows() != y.rows()       ||
         x.rows() != z.rows()          ) {
        octave_stdout << "x, y, and z must be the same size" << std::endl;
        print_usage("bfield");
        return octave_value();
    }

    Matrix Bx(x.rows(),x.columns());
    Matrix By(x.rows(),x.columns());
    Matrix Bz(x.rows(),x.columns());
    Matrix dBx(x.rows(),x.columns());
    Matrix dBy(x.rows(),x.columns());
    Matrix dBz(x.rows(),x.columns());


    for (int i = 0; i < x.columns(); i++) {
        for (int j = 0; j < x.rows(); j++) {
            Vector<double,3> r(V3(x(j,i), y(j,i), z(j,i)));
            Vector<double,3> B(0.0);
            bsrc.getb(B, r);
            Bx(j,i) = B[X];
            By(j,i) = B[Y];
            Bz(j,i) = B[Z];

            if (nargout > 3) {
                Vector<double,3> dB(0.0);
                bsrc.gradient(dB, r);
                dBx(j,i) = dB[X];
                dBy(j,i) = dB[Y];
                dBz(j,i) = dB[Z];
            }
        }/* for */
    }/* for */

    octave_value_list retval;
    retval.append(octave_value(Bx))
          .append(octave_value(By))
          .append(octave_value(Bz));

    switch (nargout) {
        case 4: retval.append(octave_value(dBx));
                break;
        case 5: retval.append(octave_value(dBx))
                      .append(octave_value(dBy));
                break;
        case 6: retval.append(octave_value(dBx))
                      .append(octave_value(dBy))
                      .append(octave_value(dBz));
                break;
        default: break;
    }

    return retval;
}
